package com.tutego.jrtf;

import java.io.*;
import java.nio.charset.Charset;
import java.util.*;

/**
 * Main class to build and stream a RTF document.
 * <p>
 * Example how to generate a RTF file:
 * <pre>
 * Rtf.document().p( "Hello", bold("RTF") ).out( new FileWriter("out.rtf") );
 * </pre>
 */
public class Rtf
{
  /** Charset used for converting chars in the range of 127 < x < 255. */
  final static Charset CHARSET1252 = Charset.forName( "Windows-1252" );
  
  /** Associates an index with an color. */
  private SortedMap<Integer, RtfHeaderColor> headerColors = new TreeMap<Integer, RtfHeaderColor>();

  /** List of fonts. */
  private List<RtfHeaderFont> headerFonts = new ArrayList<RtfHeaderFont>();

//  /** List of style sheets. */
//  private List<RtfHeaderStyle> headerStyles = new ArrayList<RtfHeaderStyle>();

  /** Document info. */
  private StringBuilder info = new StringBuilder();

  /** Document format. */
  private StringBuilder docfmt = new StringBuilder();

  /** Section data will be stored in 2 lists: One for the section formatting and headers */
  private List<CharSequence> secfmtHdrftrs = new ArrayList<CharSequence>();
  /** and another list for the paragraphs itself. */
  private List<RtfPara[]> sectionParagraphs = new ArrayList<RtfPara[]>();

  /**
   * Private constructor. The user will not instantiate this class.
   */
  private Rtf() { }

  /**
   * Creates a new RTF document.
   * @return The new RTF document.
   */
  public static Rtf rtf()
  {
    return new Rtf();
  }

  /**
   * Converts a given char sequence into RTF format and stream it to the {@code Appendable}.
   * @param out
   * @param rawText
   * @throws IOException 
   */
  static void asRtf( Appendable out, String rawText ) throws IOException
  {
    byte[] bytes = rawText.getBytes( CHARSET1252 );
    
    assert rawText.length() == bytes.length;

    for ( int i = 0; i < bytes.length; i++ )
    {
      char b = (char) (bytes[ i ] & 0xFF);

      if ( b == '\n' )
        out.append( "\\par\n" );
      else if ( b == '\t' )
        out.append( "\\tab\n" );
      else if ( b == '\\' )
        out.append( "\\\\" );
      else if ( b == '{' )
        out.append( "\\{" );
      else if ( b == '}' )
        out.append( "\\}" );
      else if ( b < 127 )
        out.append( b );
      else if ( b < 0xff )  // Use encoding 'xx if value  0x7f < char < 0xff
        out.append( "\\'" + Integer.toHexString( b ) );
      else // Use Unicode and ask the char from the String object
        out.append( "\\u" ).append( Integer.toString( rawText.charAt( i ) ) ).append( '?' );
    }
  }

  /**
   * Converts a given string to an encoded RTF sting. A new line character will be converted to \par.
   * @param rawText Raw text.
   * @return RTF encoded string.
   */
  static String asRtf( String rawText )
  {
    // TODO: Optimize

    StringBuilder result = new StringBuilder( rawText.length() * 2 );
    try
    {
      asRtf( result, rawText );
    }
    catch ( IOException e )
    {
      // If this will happen we are really in trouble
      throw new RtfException( e );
    }
    return result.toString();
  }

  /**
   * Sets RTF headers for the document.
   * @param headers Sequence of headers.
   * @return {@code this} reference.
   */
  public Rtf header( RtfHeader... headers )
  {
    for ( RtfHeader rtfHeader : headers )
    {
      if ( rtfHeader instanceof RtfHeaderColor )
        headerColors.put( ((RtfHeaderColor) rtfHeader).colorindex, (RtfHeaderColor) rtfHeader );
      else if ( rtfHeader instanceof RtfHeaderFont )
        headerFonts.add( (RtfHeaderFont) rtfHeader );
//      else if ( rtfHeader instanceof RtfHeaderStyle )
//        headerStyles.add( (RtfHeaderStyle) rtfHeader );
    }

    return this;
  }

  /**
   * Writes information group, which contains information about the document. This can include the title, author,
   * keywords, comments, and other information specific to the file. This information is for use by a
   * document-management utility.
   * @param infos RTF info objects.
   * @return {@code this} reference.
   */
  public Rtf info( RtfInfo... infos )
  {
    for ( RtfInfo rtfInfo : infos )
      info.append( rtfInfo.rtf );

    return this;
  }
  
  /**
   * Document formatting to specify the attributes of the document,
   * such as margins and footnote placement.
   * @param documentFormattings RTF document formattings.
   * @return {@code this} reference.
   */
  public Rtf documentFormatting( RtfDocfmt... documentFormattings )
  {
    for ( RtfDocfmt rtfDocfmt : documentFormattings )
      docfmt.append( rtfDocfmt.rtf );

    return this;
  }

  /**
   * Writes the paragraphs for the RTF document. Afterwards flushes all data
   * and closes the output stream.
   * @param paragraphs Parapgraphs.
   * @return {@code this} reference.
   */
  public Rtf section( RtfPara... paragraphs )
  {
    return section( null, paragraphs );
  }

  /**
   * Writes the paragraphs for the RTF document with formattings.
   * @param secfmtHdrftr Formattings. Might be {@code null}.
   * @param paragraphs  Paragraphs to be written.
   * @return {@code this} reference.
   */
  public Rtf section( RtfSectionFormatAndHeaderFooter secfmtHdrftr, RtfPara... paragraphs )
  {
    if ( paragraphs == null )
      throw new IllegalArgumentException( "There has to be atleast one paragraph in a section" );

    // First add the style
    
    if ( secfmtHdrftr != null ) 
      secfmtHdrftrs.add( secfmtHdrftr.rtf );
    else
      secfmtHdrftrs.add( null );
    
    // then the paragraphs itself to the second list

    sectionParagraphs.add( paragraphs );

    return this;
  }

  /**
   * Writes a sequence of text to. A convenience method
   * which is equal to {@code section(RtfPara.p(texts));}.
   * @param texts Text to put in paragraph.
   */
  public void p( Object... texts )
  {
    section( RtfPara.p( texts ) );
  }

  /**
   * Writes the RTF document and send the output to an {@code Appendable}.
   * This method closes the {@link Appendable} after writing if its of type
   * {@link Closeable}.
   * @param out Destination of this RTF output.
   */
  public void out( Appendable out )
  {
    if ( out == null )
      throw new IllegalArgumentException(  "Appendable is not allowed to be null" );
    
    try
    {
      writeRtfDocument( out );
    }
    catch ( IOException e )
    {
      throw new RtfException( e );
    }
    finally
    {
      if ( out instanceof Closeable )
        try { ((Closeable) out).close(); } catch ( IOException e ) { throw new RtfException(e); }
    }    
  }
  
  /**
   * Returns the RTF document as a {@code CharSequence}.
   * @return The RTF document.
   */
  public CharSequence out()
  {
    StringBuilder result = new StringBuilder( 4096 );
    out( result );
    return result;
  }

  /**
   * Returns the RTF document as a String.
   */
  @Override
  public String toString()
  {
    return out().toString();
  }

  /**
   * Opens a RTF template for later variable substitution.
   * @param inputStream Source of the RTF file.
   * @return Template object to make the substitutions on.
   */
  public static RtfTemplate template( InputStream inputStream )
  {
    return new RtfTemplate( inputStream );
  }
  
  /**
   * Writes the complete RTF document.
   */
  private void writeRtfDocument( Appendable out ) throws IOException
  {
    /*
     * <File>     := '{' <header> <document>'}'
     * <header>   := \rtf <charset> \deff? <fonttbl> <colortbl> <stylesheet>?
     * <document> := <info>? <docfmt>* <section>+
     * <section>  := <secfmt>* <hdrftr>? <para>+ ( \sect <section>)?
     */
    
    // Write <header>
    
    /*
     * <header>   := \rtf
     *               <charset> 
     *               <deffont>
     *               \deff?
     *               <fonttbl>
     *               <filetbl>?
     *               <colortbl>?
     *               <stylesheet>?
     *               <listtables>?
     *               <revtbl>?
     *               <rsidtable>?
     *               <generator>?
     */

    out.append( "{" );   // '{' <header> <document>'}'

    // The RTF version will always be 1 and the
    // character is \ansi = Windows 1252

    out.append( "\\rtf1\\ansi\\deff0" );
    
    /*
     * <fonttbl>  := '{' \fonttbl (<fontinfo> | ('{' <fontinfo> '}'))+ '}'
     */
    out.append( "\n{\\fonttbl" );
    
    if ( headerFonts.isEmpty() )
      out.append( "{\\f0 Times New Roman;}" );
    else
    {
      for ( RtfHeaderFont font : headerFonts )
        font.writeFontInfo( out );
    }
    
    out.append( '}' );
  
    /*
     * <colortbl> := '{' \colortbl <colordef>+ '}'
     */
    if ( ! headerColors.isEmpty() )
    {
      out.append( "\n{\\colortbl" );
      
      int maxColorIndex = headerColors.lastKey().intValue();

      for ( int i = 0; i <= maxColorIndex; i++ )
      {
        RtfHeaderColor color = headerColors.get( i );
        if ( color == null )
          out.append( ';' );
        else
          color.writeColordef( out );
      }
      
      out.append( '}' );
    }
  
    /*
     * <stylesheet> := '{' \ stylesheet <style>+ '}'
     */
  
//    if ( ! headerStyles.isEmpty() )
//    {
//      out.append( "\n{\\stylesheet" );
//      for ( RtfHeaderStyle style : headerStyles )
//        style.writeStyle( out );
//      
//      out.append( '}' );
//    }
//    
    out.append( '\n' );
    
    // Write <info> 
    
    if ( info.length() > 0 )
    {
      out.append( "{\\info" );
      out.append( info );
      out.append( "}\n" );      
    }

    // Write <docfmt> 
    
    if ( docfmt.length() > 0 )
      out.append( docfmt );
    
    /*
     * <document> := <info>? <docfmt>* <section>+
     * <section>  := <secfmt>* <hdrftr>? <para>+ ( \sect <section>)?
     */

    for ( int sectionCnt = 0; sectionCnt < sectionParagraphs.size(); sectionCnt++ )
    {
      RtfPara[]    paragraphs   = sectionParagraphs.get( sectionCnt );
      CharSequence secfmtHdrftr = secfmtHdrftrs.get( sectionCnt );

      // <secfmt>* <hdrftr>?
      
      if ( secfmtHdrftr != null )
        out.append( secfmtHdrftr );

      // <para>+

      for ( RtfPara rtfPara : paragraphs )
        rtfPara.rtf( out, true );
        
      // write \sect between sections but not at the end

      if ( sectionCnt != sectionParagraphs.size() - 1 )
        out.append( "\\sect\n" );
    }

    //TODO:
    //        // Place \par between paragraphs not at the end of every paragraph.
    //        // Consequence: If there is just one paragraph there isn't any \par
    //        if ( i > 0 )
    //        {
    //          // Just add \par if the one before is a paragraph too.
    //          // we don't want \par between table rows
    //          if (    paragraphs[i-1] instanceof RtfTextPara 
    //               && paragraphs[ i ] instanceof RtfTextPara )
    //            out.append( "\\par\n" );
    //        }

    // We are done

    out.append( "}" );
  }
}
