package com.tutego.jrtf;

/**
 * Class to set the document format like page width, margins.
 */
public class RtfDocfmt
{
  /** RTF with format string. */
  String rtf;

  /**
   * Initialize this instance with the RTF control word.
   * @param rtf
   */
  private RtfDocfmt( String rtf )
  {
    this.rtf = rtf;
  }

  /**
   * Default tab width.
   * @param width  Width of the tab.
   * @param unit   Unit of the tab.
   * @return New RtfDocfmt object.
   */
  public static RtfDocfmt defaultTab( double width, RtfUnit unit )
  {
    return new RtfDocfmt( "\\deftab" + unit.toTwips( width ));
  }

  /**
   * Footnotes only or nothing at all (default).
   * @return New RtfDocfmt object.
   */
  public static RtfDocfmt footnotesOnly()
  {
    return new RtfDocfmt( "\\fet0" );
  }

  // Footnotes and endnotes
  
  /**
   * Endnotes only.
   * @return New RtfDocfmt object.
   */
  public static RtfDocfmt endnotesOnly()
  {
    return new RtfDocfmt( "\\fet1" );
  }

  /**
   * Footnotes and endnotes both.
   * @return New RtfDocfmt object.
   */
  public static RtfDocfmt footnotesAndEndnotes()
  {
    return new RtfDocfmt( "\\fet2" );
  }

  /**
   * Footnote numbering - Arabic numbering (1, 2, 3, _).
   * @return New RtfDocfmt object.
   */
  public static RtfDocfmt footnoteNumberingArabic()
  {
    return new RtfDocfmt( "\\ftnnar" );
  }

  /**
   * Footnote numbering - Alphabetic uppercase (A, B, C, _).
   * @return New RtfDocfmt object.
   */
  public static RtfDocfmt footnoteNumberingUpperAlphabetic()
  {
    return new RtfDocfmt( "\\ftnnauc" );
  }

  /**
   * Footnote numbering - Roman lowercase (i, ii, iii, _).
   * @return New RtfDocfmt object.
   */
  public static RtfDocfmt footnoteNumberingUpperRoman()
  {
    return new RtfDocfmt( "\\ftnnrlc" );
  }

  // Page information

  /**
   * Paper width. Default is 12,240 twips.
   * @param width  Width of the page.
   * @param unit   Measurement unit.
   * @return New RtfDocfmt object.
   */
  public static RtfDocfmt paperWidth( double width, RtfUnit unit)
  {
    return new RtfDocfmt( "\\paperw" + unit.toTwips( width ) );
  }

  /**
   * Paper height. Default is 15,840 twips.
   * @param height Height of the page.
   * @param unit   Measurement unit.
   * @return New RtfDocfmt object.
   */
  public static RtfDocfmt paperHeight( double height, RtfUnit unit)
  {
    return new RtfDocfmt( "\\paperh" + unit.toTwips( height) );
  }

  /**
   * Left margin. Default is 1,800 twips.
   * @param margin Left margin.
   * @param unit Measurement unit.
   * @return New RtfDocfmt object.
   */
  public static RtfDocfmt leftMargin( double margin, RtfUnit unit)
  {
    return new RtfDocfmt( "\\margl" + unit.toTwips( margin ) );
  }

  /**
   * Right margin. Default is 1,800 twips.
   * @param margin Right margin.
   * @param unit Measurement unit.
   * @return New RtfDocfmt object.
   */
  public static RtfDocfmt rightMargin( double margin, RtfUnit unit)
  {
    return new RtfDocfmt( "\\margr" + unit.toTwips( margin ) );
  }

  /**
   * Top margin. Default is 1,440 twips.
   * @param margin Top margin.
   * @param unit Measurement unit.
   * @return New RtfDocfmt object.
   */
  public static RtfDocfmt topMargin( double margin, RtfUnit unit)
  {
    return new RtfDocfmt( "\\margt" + unit.toTwips( margin ) );
  }

  /**
   * Bottom margin. Default is 1,440 twips.
   * @param margin Bottom margin.
   * @param unit Measurement unit.
   * @return New RtfDocfmt object.
   */
  public static RtfDocfmt bottomMargin( double margin, RtfUnit unit)
  {
    return new RtfDocfmt( "\\margb" + unit.toTwips( margin ) );
  }

  /**
   * Landscape format.
   * @return New RtfDocfmt object.
   */
  public static RtfDocfmt landscape()
  {
    return new RtfDocfmt( "\\landscape" );
  }
  
  // Revision marks

  /**
   * This document is protected for revisions. The user can edit the document, but revision marking cannot be
   * disabled.
   * @return New RtfDocfmt object.
   */
  public static RtfDocfmt revisionProtected()
  {
    return new RtfDocfmt( "\\revprot" );
  }

  /**
   * Turns on revision marking.
   * @return New RtfDocfmt object.
   */
  public static RtfDocfmt revisionMarking()
  {
    return new RtfDocfmt( "\\revisions" );
  }
}
