package com.tutego.jrtf;

import java.io.IOException;

/**
 * Section-Formatting.
 */
public class RtfSectionFormatAndHeaderFooter
{
  /** Holds the RTF text for a header or footer element. */
  final CharSequence rtf;

  /**
   * Initializes this object with RTF text.
   * @param rtf  RTF.
   */
  private RtfSectionFormatAndHeaderFooter( CharSequence rtf )
  {
    this.rtf = rtf;
  }

  /*
   * <section>  := <secfmt>* <hdrftr>? <para>+ ( \sect <section>)?
   */

  /**
   * Builds a RtfSectionFormatAndHeaderFooter objects representing all the
   * section formattings.
   * @param sectionFormats  Sequence of section formats.
   * @return RtfSectionFormatAndHeaderFooter
   */
  public static RtfSectionFormatAndHeaderFooter sectionFormatting( RtfSectionFormatAndHeaderFooter ... sectionFormats )
  {
    StringBuilder sb = new StringBuilder( sectionFormats.length * 10 );
    for ( RtfSectionFormatAndHeaderFooter rtfSectionFormat : sectionFormats )
      sb.append( rtfSectionFormat.rtf );

    return new RtfSectionFormatAndHeaderFooter( sb );
  }

  /**
   * Reset to default section properties.
   * @return New {@code RtfSectionFormatAndHeaderFooter} object.
   */
  public static RtfSectionFormatAndHeaderFooter reset()
  {
    return new RtfSectionFormatAndHeaderFooter( "\\sectd" );
  }

  /**
   * Endnotes included in the section.
   * @return New {@code RtfSectionFormatAndHeaderFooter} object.
   */
  public static RtfSectionFormatAndHeaderFooter endnotesIncluded()
  {
    return new RtfSectionFormatAndHeaderFooter( "\\endnhere" );
  }

  // Section break
  
  /**
   * No section break.
   * @return New {@code RtfSectionFormatAndHeaderFooter} object.
   */
  public static RtfSectionFormatAndHeaderFooter noBreak()
  {
    return new RtfSectionFormatAndHeaderFooter( "\\sbknone" );
  }
  
  /**
   * Section break starts a new column.
   * @return New {@code RtfSectionFormatAndHeaderFooter} object.
   */
  public static RtfSectionFormatAndHeaderFooter breakStartsNewColumn()
  {
    return new RtfSectionFormatAndHeaderFooter( "\\sbkcol" );
  }

  /**
   * Section break starts a new page. Thats the default.
   * @return New {@code RtfSectionFormatAndHeaderFooter} object.
   */
  public static RtfSectionFormatAndHeaderFooter breakStartsNewPage()
  {
    return new RtfSectionFormatAndHeaderFooter( "\\sbkpage" );
  }

  /**
   * Section break starts at an even page.
   * @return New {@code RtfSectionFormatAndHeaderFooter} object.
   */
  public static RtfSectionFormatAndHeaderFooter breakStartsNewEvenPage()
  {
    return new RtfSectionFormatAndHeaderFooter( "\\sbkeven" );
  }

  /**
   * Section break starts at an odd page.
   * @return New {@code RtfSectionFormatAndHeaderFooter} object.
   */
  public static RtfSectionFormatAndHeaderFooter breakStartsNewOddPage()
  {
    return new RtfSectionFormatAndHeaderFooter( "\\sbkodd" );
  }

  // Columns

  /**
   * Number of columns. Default is 1.
   * @param columns Number of columns for this section.
   * @return New {@code RtfSectionFormatAndHeaderFooter} object.
   */
  public static RtfSectionFormatAndHeaderFooter columns( int columns )
  {
    if ( columns <= 0 )
      throw new RtfException( "Number of colums can't be <= 0" );

    return new RtfSectionFormatAndHeaderFooter( "\\cols" + columns );
  }

  /**
   * Space between columns. Default is 720 twips.
   * @param space Space between columns.
   * @param unit Measurement.
   * @return New {@code RtfSectionFormatAndHeaderFooter} object.
   */
  public static RtfSectionFormatAndHeaderFooter spaceBetweenColumns( double space, RtfUnit unit )
  {
    return new RtfSectionFormatAndHeaderFooter( "\\colsx" + unit.toTwips( space ));
  }

  /**
   * Line between columns.
   * @return New {@code RtfSectionFormatAndHeaderFooter} object.
   */
  public static RtfSectionFormatAndHeaderFooter lineBetweenColumns()
  {
    return new RtfSectionFormatAndHeaderFooter( "\\linebetcol" );
  }
  
  // Line numbering

  // Page information
  
  // Page numbers
  
  /**
   * Beginning page number. Default is 1.
   * @param pageNumber Beginning page number.
   * @return New {@code RtfSectionFormatAndHeaderFooter} object.
   */
  public static RtfSectionFormatAndHeaderFooter beginningPageNumber( int pageNumber )
  {
    return new RtfSectionFormatAndHeaderFooter( "\\pgnstarts" + pageNumber );
  }

  // Vertical alignment

  /**
   * Text is top-aligned. This is the default.
   * @return New {@code RtfSectionFormatAndHeaderFooter} object.
   */
  public static RtfSectionFormatAndHeaderFooter topAlignText()
  {
    return new RtfSectionFormatAndHeaderFooter( "\\vertalt" );
  }

  /**
   * Text is bottom-aligned.
   * @return New {@code RtfSectionFormatAndHeaderFooter} object.
   */
  public static RtfSectionFormatAndHeaderFooter bottomAlignText()
  {
    return new RtfSectionFormatAndHeaderFooter( "\\vertalb" );
  }

  /**
   * Text is centered vertically.
   * @return New {@code RtfSectionFormatAndHeaderFooter} object.
   */
  public static RtfSectionFormatAndHeaderFooter centerVerticalText()
  {
    return new RtfSectionFormatAndHeaderFooter( "\\vertalc" );
  }

  /**
   * Text is justified vertically.
   * @return New {@code RtfSectionFormatAndHeaderFooter} object.
   */
  public static RtfSectionFormatAndHeaderFooter justifyVerticalText()
  {
    return new RtfSectionFormatAndHeaderFooter( "\\vertalj" );
  }

  
  /*
   * <section> := <secfmt>* <hdrftr>? <para>+ ( \sect <section>)?
   * <hdrftr>  := '{' <hdrctl> <para>+ '}' <hdrftr>?
   * <hdrctl>  := \header | \footer | \headerl | \headerr | \headerf | \footerl | \ footerr | \footerf
   */
 
  /**
   * Puts a header with a given paragraph on all pages.
   * @param para Paragraph for the header.
   * @return New {@code RtfSectionFormatAndHeaderFooter} object.
   */
  public static RtfSectionFormatAndHeaderFooter headerOnAllPages( RtfPara para )
  {
    try
    {
      StringBuilder out = new StringBuilder( "{\\header" );
      para.rtf( out, false ) ;
      out.append( '}' );
      return new RtfSectionFormatAndHeaderFooter( out );
    }
    catch ( IOException e )
    {
      throw new RtfException( e );
    }
  }
}
