package com.tutego.jrtf;

import java.io.*;
import java.util.*;
import java.util.regex.*;

/**
 * This class is able to make variable substitutions in an
 * existing RTF file.
 * <pre>
 * Rtf.template( new FileInputStream("template.rtf) )
 *    .inject( "ADDRESSLINE1", "tutego" )
 *    .inject( "ADDRESSLINE2", bold("Sonsbeck") )
 *    .out( FileOutputStream("out.rtf") );
 * </pre>
 */
public class RtfTemplate
{
  /** Holds the template file. */
  private StringBuilder template = new StringBuilder( 8192 );

  /** Map with all variables and substitutions. */
  private Map<String, Object> map = new HashMap<String, Object>();

  /** Regex pattern for %%VARIABLE%%. */
  private final static Pattern VARIABLEPATTERN = Pattern.compile( "%%(\\S+)%%", Pattern.DOTALL | Pattern.MULTILINE);

  /**
   * Reads the template from that {@link InputStream}. The method treats the
   * byte in Windows-1252 encoding. After reading the method closes the 
   * stream.
   * @param inputStream
   */
  RtfTemplate( InputStream inputStream )
  {
    Reader reader = null;
    try
    {
      if ( ! ( inputStream instanceof BufferedInputStream ) )
        inputStream = new BufferedInputStream( inputStream );

      reader = new InputStreamReader( inputStream, Rtf.CHARSET1252 );      

      for ( int c; (c = reader.read() )!= -1; )
        template.append( (char ) c );
    }
    catch ( IOException e )
    {
      throw new RtfException( e );
    }
    finally
    {
      if ( reader != null )
        try { reader.close(); } catch ( IOException e ) { throw new RtfException( e ); }
    }
  }

  /**
   * Adds key/values pairs for a variable substitution.
   * @param map Map with key/value pairs.
   * @return {@code this} object.
   */
  public RtfTemplate inject( Map<String, Object> map )
  {
    this.map.putAll( map );  
    return this;
  }

  /**
   * Adds a key/value pair for substitution.
   * @param key Key.
   * @param value Value.
   * @return {@code this} object.
   */
  public RtfTemplate inject( String key, String value )
  {
    map.put( key, value );
    return this;
  }

  /**
   * Performs the variable transformation and returns the
   * transformed RTF document.
   * @return RTF document after variable substitution.
   */
  public String out()
  {
    if ( map.isEmpty() )
      return template.toString();

    StringBuffer result = new StringBuffer( template.length() );
    Matcher matcher = VARIABLEPATTERN.matcher( template );
  
    while ( matcher.find() )
    {
      Object value = map.get( matcher.group( 1 ) );
  
      if ( value == null )
        continue;
  
      StringBuilder sb = new StringBuilder( 128 );
      try { RtfText.text( value ).rtf( sb ); }
      catch ( IOException e ) {
        throw new RtfException( e );
      }
      matcher.appendReplacement( result, Matcher.quoteReplacement( sb.toString() ) );
    }

    matcher.appendTail( result );
  
    return result.toString();
  }

  /**
   * Performs the variable transformation and writes the RTF document and send the output to an {@code
   * Appendable}. This method closes the {@link Appendable} after writing if its of type {@link Closeable}.
   * @param out Destination of this RTF output.
   */
  public void out( OutputStream out )
  {
    try
    {
      String out2 = out();
      out.write( out2.getBytes(Rtf.CHARSET1252) );
    }
    catch ( IOException e )
    {
      throw new RtfException( e );
    }
    finally
    {
      if ( out instanceof Closeable )
        try { ((Closeable) out).close(); } catch ( IOException e ) { throw new RtfException(e); }
    }
  }
}
